# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok import Declarations
from anyblok.config import Configuration
from anyblok.field import FieldException
from anyblok.column import Integer, String
from anyblok.relationship import One2One
from sqlalchemy import ForeignKeyConstraint
from .conftest import init_registry, reset_db

register = Declarations.register
Model = Declarations.Model


@pytest.fixture(
    scope="class",
    params=[
        ('prefix', 'suffix'),
        ('', ''),
    ]
)
def db_schema(request, bloks_loaded):
    Configuration.set('prefix_db_schema', request.param[0])
    Configuration.set('suffix_db_schema', request.param[1])

    def rollback():
        Configuration.set('prefix_db_schema', '')
        Configuration.set('suffix_db_schema', '')

    request.addfinalizer(rollback)


def _minimum_one2one(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = One2One(model=Model.Address, backref="person")


def _minimum_one2one_with_schema(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2m_schema'

        id = Integer(primary_key=True)

    @register(Model)
    class Person:
        __db_schema__ = 'test_db_m2m_schema'

        name = String(primary_key=True)
        address = One2One(model=Model.Address, backref="person")


def _minimum_one2one_with_diferent_schema1(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2m_schema'

        id = Integer(primary_key=True)

    @register(Model)
    class Person:
        __db_schema__ = 'test_other_db_schema'

        name = String(primary_key=True)
        address = One2One(model=Model.Address, backref="person")


def _minimum_one2one_with_diferent_schema2(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2m_schema'

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = One2One(model=Model.Address, backref="person")


def _one2one_with_str_method(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = One2One(model="Model.Address", backref="person")


def _minimum_one2one_on_mapper_1(**kwargs):

    @register(Model, tablename="x")
    class Address:

        id = Integer(primary_key=True, db_column_name="x1")

    @register(Model, tablename="y")
    class Person:

        name = String(primary_key=True, db_column_name="y1")
        address = One2One(model=Model.Address, backref="person")


def _minimum_one2one_on_mapper_2(**kwargs):

    @register(Model, tablename="x")
    class Address:

        id = Integer(primary_key=True, db_column_name="x1")

    @register(Model, tablename="y")
    class Person:

        name = String(primary_key=True, db_column_name="y1")
        address_id = Integer(db_column_name="y2",
                             foreign_key=Model.Address.use('id'))
        address = One2One(model=Model.Address, backref="person")


@pytest.fixture(scope="class", params=[
    _minimum_one2one,
    _minimum_one2one_with_schema,
    _minimum_one2one_with_diferent_schema1,
    _minimum_one2one_with_diferent_schema2,
    _one2one_with_str_method,
    _minimum_one2one_on_mapper_1,
    _minimum_one2one_on_mapper_2,
])
def registry_minimum_one2one(request, bloks_loaded, db_schema):
    reset_db()
    registry = init_registry(request.param)
    request.addfinalizer(registry.close)
    return registry


@pytest.mark.relationship
class TestMinimumOne2One:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_minimum_one2one):
        transaction = registry_minimum_one2one.begin_nested()
        request.addfinalizer(transaction.rollback)

    def test_minimum_one2one(self, registry_minimum_one2one):
        registry = registry_minimum_one2one
        assert hasattr(registry.Person, 'address')
        assert hasattr(registry.Person, 'address_id')
        address = registry.Address.insert()
        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)

        assert address.person is person

    def test_one2one_multi_entry_for_same(self, registry_minimum_one2one):
        registry = registry_minimum_one2one
        address = registry.Address.insert()
        p1 = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)
        assert p1.address is address
        p2 = registry.Person.insert(name="Franck BRET", address=address)
        assert p2.address is address
        assert p1.address_id is None

    def test_minimum_one2one_expire_field(self, registry_minimum_one2one):
        registry = registry_minimum_one2one
        assert ('address',) in registry.expire_attributes[
            'Model.Person']['address_id']
        assert ('address', 'person') in registry.expire_attributes[
            'Model.Person']['address_id']


def _multi_fk_one2one():

    @register(Model)
    class Test:

        id = Integer(primary_key=True, unique=True)
        id2 = String(primary_key=True, unique=True)

    @register(Model)
    class Test2:

        id = Integer(primary_key=True)
        test = One2One(model=Model.Test,
                       column_names=('test_id', 'test_id2'),
                       backref="test2")


@pytest.fixture(scope="class")
def registry_multi_fk_one2one(request, bloks_loaded):
    reset_db()
    registry = init_registry(_multi_fk_one2one)
    request.addfinalizer(registry.close)
    return registry


@pytest.mark.relationship
class TestMultiFKOne2One:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_multi_fk_one2one):
        transaction = registry_multi_fk_one2one.begin_nested()
        request.addfinalizer(transaction.rollback)

    def test_with_multi_foreign_key_on_unexisting_named_column(
        self, registry_multi_fk_one2one
    ):
        registry = registry_multi_fk_one2one
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        assert test.id == test2.test_id
        assert test.id2 == test2.test_id2

    def test_multi_o2o_one2one_expire_field(
        self, registry_multi_fk_one2one
    ):
        registry = registry_multi_fk_one2one
        assert ('test',) in registry.expire_attributes[
            'Model.Test2']['test_id']
        assert ('test', 'test2') in registry.expire_attributes[
            'Model.Test2']['test_id']
        assert ('test',) in registry.expire_attributes[
            'Model.Test2']['test_id2']
        assert ('test', 'test2') in registry.expire_attributes[
            'Model.Test2']['test_id2']


def _complete_one2one(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = One2One(model=Model.Address,
                          remote_columns="id", column_names="id_of_address",
                          backref="person", nullable=False)


def _complete_one2one_with_schema(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2m_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:
        __db_schema__ = 'test_db_m2m_schema'

        name = String(primary_key=True)
        address = One2One(model=Model.Address,
                          remote_columns="id", column_names="id_of_address",
                          backref="person", nullable=False)


def _minimum_one2one_without_backref(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = One2One(model=Model.Address)


def _minimum_one2one_with_one2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address = One2One(model=Model.Address,
                          backref="person", one2many="persons")


@pytest.mark.relationship
class TestOne2One:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        reset_db()
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_complete_one2one(self):
        registry = self.init_registry(_complete_one2one)
        assert hasattr(registry.Person, 'address')
        assert hasattr(registry.Person, 'id_of_address')
        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')
        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)

        assert address.person is person

    def test_complete_one2one_with_schema(self, db_schema):
        registry = self.init_registry(_complete_one2one_with_schema)
        assert hasattr(registry.Person, 'address')
        assert hasattr(registry.Person, 'id_of_address')
        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')
        person = registry.Person.insert(
            name="Jean-sébastien SUZANNE", address=address)

        assert address.person is person

    def test_minimum_one2one_without_backref(self):
        with pytest.raises(FieldException):
            self.init_registry(_minimum_one2one_without_backref)

    def test_minimum_one2one_with_one2many(self):
        with pytest.raises(FieldException):
            self.init_registry(_minimum_one2one_with_one2many)

    def test_complet_with_multi_foreign_key(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                @classmethod
                def define_table_args(cls):
                    table_args = super(Test2, cls).define_table_args()
                    return table_args + (ForeignKeyConstraint(
                        [cls.test_id, cls.test_id2], ['test.id', 'test.id2']),)

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = One2One(model=Model.Test,
                               remote_columns=('id', 'id2'),
                               column_names=('test_id', 'test_id2'),
                               backref="test2")

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        assert test.id == test2.test_id
        assert test.id2 == test2.test_id2

    def test_complet_with_multi_foreign_key_without_constraint(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = One2One(model=Model.Test,
                               remote_columns=('id', 'id2'),
                               column_names=('test_id', 'test_id2'),
                               backref="test2")

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        assert test.id == test2.test_id
        assert test.id2 == test2.test_id2

    def test_with_multi_foreign_key_on_existing_column(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = One2One(model=Model.Test, backref="test2")

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        assert test.id == test2.test_id
        assert test.id2 == test2.test_id2

    def test_with_multi_foreign_key_on_existing_column2(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                other_test_id = Integer(
                    foreign_key=Model.Test.use('id'), nullable=False)
                other_test_id2 = String(
                    foreign_key=Model.Test.use('id2'), nullable=False)
                test = One2One(model=Model.Test, backref="test2")

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        assert test.id == test2.other_test_id
        assert test.id2 == test2.other_test_id2

    def test_with_multi_foreign_key_on_unexisting_column(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test = One2One(model=Model.Test, backref="test2")

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(id2="10")
        test2 = registry.Test2.insert(test=test)
        assert test.id == test2.test_id
        assert test.id2 == test2.test_id2

    def test_with_multi_foreign_key_on_unexisting_named_column2(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = Integer(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test = One2One(model=Model.Test, column_names=(
                    'other_test_id', 'other_test_id2'), backref="test2")

        with pytest.raises(FieldException):
            self.init_registry(add_in_registry)



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#    Copyright (C) 2002 Pierre Verkest <pierreverkest84@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.testing import sgdb_in
from anyblok.blok import BlokManager
from anyblok.registry import RegistryException, RegistryConflictingException


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                    reason='Not for MySQL and MariaDB')
class TestBlok:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok):
        transaction = registry_testblok.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_blok_exist(self, registry_testblok):
        registry = registry_testblok
        Blok = registry.System.Blok
        query = Blok.query().filter(Blok.name == 'test-blok1')
        if not query.count():
            pytest.fail('No blok found')

        testblok1 = query.first()
        assert testblok1.state == 'uninstalled'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version is None
        assert testblok1.short_description == ''
        assert testblok1.long_description == ''

    def test_install(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'

    def test_install_an_installed_blok(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok1',))
        with pytest.raises(RegistryException):
            registry.upgrade(install=('test-blok1',))

    def test_uninstall(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok1',))
        registry.upgrade(uninstall=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        assert testblok1.state == 'uninstalled'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version is None

    def test_uninstall_an_uninstalled_blok(self, registry_testblok):
        registry = registry_testblok
        with pytest.raises(RegistryException):
            registry.upgrade(uninstall=('test-blok1',))

    def test_update(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok1.version = '2.0.0'
        registry.upgrade(update=('test-blok1',))
        assert testblok1.state == 'installed'
        assert testblok1.installed_version == '2.0.0'

    def test_update_an_uninstalled_blok(self, registry_testblok):
        registry = registry_testblok
        with pytest.raises(RegistryException):
            registry.upgrade(update=('test-blok1',))


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                    reason='Not for MySQL and MariaDB')
class TestBlokRequired:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok):
        transaction = registry_testblok.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_blok_exist(self, registry_testblok):
        registry = registry_testblok
        Blok = registry.System.Blok
        query = Blok.query().filter(Blok.name == 'test-blok2')
        if not query.count():
            pytest.fail('No blok found')

        testblok2 = query.first()
        assert testblok2.state == 'uninstalled'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version is None
        assert testblok2.short_description == 'Test blok2'

    def test_install_1by1(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok2.state == 'uninstalled'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version is None
        registry.upgrade(install=('test-blok2',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok2.state == 'installed'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version == '1.0.0'

    def test_install(self, registry_testblok):
        registry = registry_testblok
        Blok = registry.System.Blok
        registry.upgrade(install=('test-blok2',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok2.state == 'installed'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version == '1.0.0'

    def test_uninstall(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok2',))
        registry.upgrade(uninstall=('test-blok2',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok2.state == 'uninstalled'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version is None

    def test_uninstall_required(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok2',))
        registry.upgrade(uninstall=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        assert testblok1.state == 'uninstalled'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version is None
        assert testblok2.state == 'uninstalled'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version is None

    def test_update(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok2',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok1.version = '2.0.0'
        testblok2.version = '2.0.0'
        registry.upgrade(update=('test-blok2',))
        assert testblok1.state == 'installed'
        assert testblok1.installed_version == '1.0.0'
        assert testblok2.state == 'installed'
        assert testblok2.installed_version == '2.0.0'

    def test_update_required(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok2',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok1.version = '2.0.0'
        testblok2.version = '2.0.0'
        registry.upgrade(update=('test-blok1',))
        assert testblok1.state == 'installed'
        assert testblok1.installed_version == '2.0.0'
        assert testblok2.state == 'installed'
        assert testblok2.installed_version == '2.0.0'


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                    reason='Not for MySQL and MariaDB')
class TestBlokRequired2:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok):
        transaction = registry_testblok.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_blok_exist(self, registry_testblok):
        registry = registry_testblok
        Blok = registry.System.Blok
        query = Blok.query().filter(Blok.name == 'test-blok3')
        if not query.count():
            pytest.fail('No blok found')

        testblok2 = query.first()
        assert testblok2.state == 'uninstalled'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version is None
        long_description = ".. This file is a part of the AnyBlok project\n.."
        long_description += "\n..    Copyright (C) 2014 Jean-Sebastien SUZANNE"
        long_description += " <jssuzanne@anybox.fr>\n..\n.. This Source Code "
        long_description += "Form is subject to the terms of the Mozilla "
        long_description += "Public License,\n.. v. 2.0. If a copy of the MPL "
        long_description += "was not distributed with this file,You can\n.. "
        long_description += "obtain one at http://mozilla.org/MPL/2.0/.\n\n"
        long_description += "Test blok3\n"
        assert testblok2.long_description == long_description

    def test_install_1by1(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok2.state == 'uninstalled'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version is None
        assert testblok3.state == 'uninstalled'
        assert testblok3.version == '1.0.0'
        assert testblok3.installed_version is None
        registry.upgrade(install=('test-blok2',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok2.state == 'installed'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version == '1.0.0'
        assert testblok3.state == 'uninstalled'
        assert testblok3.version == '1.0.0'
        assert testblok3.installed_version is None
        registry.upgrade(install=('test-blok3',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok2.state == 'installed'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version == '1.0.0'
        assert testblok3.state == 'installed'
        assert testblok3.version == '1.0.0'
        assert testblok3.installed_version == '1.0.0'

    def test_install(self, registry_testblok):
        registry = registry_testblok
        Blok = registry.System.Blok
        registry.upgrade(install=('test-blok3',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok2.state == 'installed'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version == '1.0.0'
        assert testblok3.state == 'installed'
        assert testblok3.version == '1.0.0'
        assert testblok3.installed_version == '1.0.0'

    def test_uninstall(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok3',))
        registry.upgrade(uninstall=('test-blok3',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        assert testblok1.state, 'installed'
        assert testblok1.version, '1.0.0'
        assert testblok1.installed_version, '1.0.0'
        assert testblok2.state, 'installed'
        assert testblok2.version, '1.0.0'
        assert testblok2.installed_version, '1.0.0'
        assert testblok3.state, 'uninstalled'
        assert testblok3.version, '1.0.0'
        assert testblok3.installed_version is None

    def test_uninstall_first_required(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok3',))
        registry.upgrade(uninstall=('test-blok2',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok2.state == 'uninstalled'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version is None
        assert testblok3.state == 'uninstalled'
        assert testblok3.version == '1.0.0'
        assert testblok3.installed_version is None

    def test_uninstall_all_required(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok3',))
        registry.upgrade(uninstall=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        assert testblok1.state == 'uninstalled'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version is None
        assert testblok2.state == 'uninstalled'
        assert testblok2.version == '1.0.0'
        assert testblok2.installed_version is None
        assert testblok3.state == 'uninstalled'
        assert testblok3.version == '1.0.0'
        assert testblok3.installed_version is None

    def test_update(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok3',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        testblok1.version = '2.0.0'
        testblok2.version = '2.0.0'
        testblok3.version = '2.0.0'
        registry.upgrade(update=('test-blok3',))
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        assert testblok1.state == 'installed'
        assert testblok1.installed_version == '1.0.0'
        assert testblok2.state == 'installed'
        assert testblok2.installed_version == '1.0.0'
        assert testblok3.state == 'installed'
        assert testblok3.installed_version == '2.0.0'

    def test_update_first_required(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok3',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        testblok1.version = '2.0.0'
        testblok2.version = '2.0.0'
        testblok3.version = '2.0.0'
        registry.upgrade(update=('test-blok2',))
        assert testblok1.state == 'installed'
        assert testblok1.installed_version == '1.0.0'
        assert testblok2.state == 'installed'
        assert testblok2.installed_version == '2.0.0'
        assert testblok3.state == 'installed'
        assert testblok3.installed_version == '2.0.0'

    def test_update_all_required(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok3',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok2 = Blok.query().filter(Blok.name == 'test-blok2').first()
        testblok3 = Blok.query().filter(Blok.name == 'test-blok3').first()
        testblok1.version = '2.0.0'
        testblok2.version = '2.0.0'
        testblok3.version = '2.0.0'
        registry.upgrade(update=('test-blok1',))
        assert testblok1.state == 'installed'
        assert testblok1.installed_version == '2.0.0'
        assert testblok2.state == 'installed'
        assert testblok2.installed_version == '2.0.0'
        assert testblok3.state == 'installed'
        assert testblok3.installed_version == '2.0.0'


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                    reason='Not for MySQL and MariaDB')
class TestBlokConditionnal:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok):
        transaction = registry_testblok.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_install_1by1(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok4 = Blok.query().filter(Blok.name == 'test-blok4').first()
        testblok5 = Blok.query().filter(Blok.name == 'test-blok5').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok4.state == 'uninstalled'
        assert testblok4.version == '1.0.0'
        assert testblok4.installed_version is None
        assert testblok5.state == 'uninstalled'
        assert testblok5.version == '1.0.0'
        assert testblok5.installed_version is None
        registry.upgrade(install=('test-blok4',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok4 = Blok.query().filter(Blok.name == 'test-blok4').first()
        testblok5 = Blok.query().filter(Blok.name == 'test-blok5').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok4.state == 'installed'
        assert testblok4.version == '1.0.0'
        assert testblok4.installed_version == '1.0.0'
        assert testblok5.state == 'installed'
        assert testblok5.version == '1.0.0'
        assert testblok5.installed_version == '1.0.0'

    def test_install(self, registry_testblok):
        registry = registry_testblok
        Blok = registry.System.Blok
        registry.upgrade(install=('test-blok5',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok4 = Blok.query().filter(Blok.name == 'test-blok4').first()
        testblok5 = Blok.query().filter(Blok.name == 'test-blok5').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok4.state == 'installed'
        assert testblok4.version == '1.0.0'
        assert testblok4.installed_version, '1.0.0'
        assert testblok5.state == 'installed'
        assert testblok5.version == '1.0.0'
        assert testblok5.installed_version == '1.0.0'

    def test_uninstall(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok5',))
        with pytest.raises(RegistryException):
            registry.upgrade(uninstall=('test-blok5',))

    def test_uninstall_conditionnal(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok5',))
        registry.upgrade(uninstall=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok4 = Blok.query().filter(Blok.name == 'test-blok4').first()
        testblok5 = Blok.query().filter(Blok.name == 'test-blok5').first()
        assert testblok1.state, 'uninstalled'
        assert testblok1.version, '1.0.0'
        assert testblok1.installed_version is None
        assert testblok4.state, 'installed'
        assert testblok4.version, '1.0.0'
        assert testblok4.installed_version, '1.0.0'
        assert testblok5.state, 'uninstalled'
        assert testblok5.version, '1.0.0'
        assert testblok5.installed_version is None

    def test_update(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok5',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok4 = Blok.query().filter(Blok.name == 'test-blok4').first()
        testblok5 = Blok.query().filter(Blok.name == 'test-blok5').first()
        testblok1.version = '2.0.0'
        testblok4.version = '2.0.0'
        testblok5.version = '2.0.0'
        registry.upgrade(update=('test-blok5',))
        assert testblok1.state == 'installed'
        assert testblok1.installed_version == '1.0.0'
        assert testblok4.state == 'installed'
        assert testblok4.installed_version == '1.0.0'
        assert testblok5.state == 'installed'
        assert testblok5.installed_version == '2.0.0'

    def test_update_conditionnal(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok5',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok4 = Blok.query().filter(Blok.name == 'test-blok4').first()
        testblok5 = Blok.query().filter(Blok.name == 'test-blok5').first()
        testblok1.version = '2.0.0'
        testblok4.version = '2.0.0'
        testblok5.version = '2.0.0'
        registry.upgrade(update=('test-blok1',))
        assert testblok1.state == 'installed'
        assert testblok1.installed_version == '2.0.0'
        assert testblok4.state == 'installed'
        assert testblok4.installed_version == '1.0.0'
        assert testblok5.state == 'installed'
        assert testblok5.installed_version == '2.0.0'


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                    reason='Not for MySQL and MariaDB')
class TestBlokOptional:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok):
        transaction = registry_testblok.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_install_1by1(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok1',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok6.state == 'uninstalled'
        assert testblok6.version == '1.0.0'
        assert testblok6.installed_version is None
        registry.upgrade(install=('test-blok6',))
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok6.state == 'installed'
        assert testblok6.version == '1.0.0'
        assert testblok6.installed_version == '1.0.0'

    def test_install(self, registry_testblok):
        registry = registry_testblok
        Blok = registry.System.Blok
        registry.upgrade(install=('test-blok6',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok6.state == 'installed'
        assert testblok6.version == '1.0.0'
        assert testblok6.installed_version == '1.0.0'

    def test_uninstall(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok6',))
        registry.upgrade(uninstall=('test-blok6',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        assert testblok1.state == 'installed'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version == '1.0.0'
        assert testblok6.state == 'uninstalled'
        assert testblok6.version == '1.0.0'
        assert testblok6.installed_version is None

    def test_uninstall_optional(self, registry_testblok):
        registry = registry_testblok
        Blok = registry.System.Blok
        registry.upgrade(install=('test-blok6',))
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        testblok6.version = '2.0.0'
        registry.upgrade(uninstall=('test-blok1',))
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        assert testblok1.state == 'uninstalled'
        assert testblok1.version == '1.0.0'
        assert testblok1.installed_version is None
        assert testblok6.state == 'installed'
        assert testblok6.version == '2.0.0'
        assert testblok6.installed_version == '2.0.0'

    def test_update(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok6',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        testblok1.version = '2.0.0'
        testblok6.version = '2.0.0'
        registry.upgrade(update=('test-blok6',))
        assert testblok1.state == 'installed'
        assert testblok1.installed_version == '1.0.0'
        assert testblok6.state == 'installed'
        assert testblok6.installed_version == '2.0.0'

    def test_update_optional(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok6',))
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok6 = Blok.query().filter(Blok.name == 'test-blok6').first()
        testblok1.version = '2.0.0'
        testblok6.version = '2.0.0'
        registry.upgrade(update=('test-blok1',))
        assert testblok1.state == 'installed'
        assert testblok1.installed_version == '2.0.0'
        assert testblok6.state == 'installed'
        assert testblok6.installed_version == '2.0.0'


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                    reason='Not for MySQL and MariaDB')
class TestBlokConflicting:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok):
        transaction = registry_testblok.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_marked_as_conflicting(self, registry_testblok):
        blok1 = BlokManager.get('test-blok1')
        blok13 = BlokManager.get('test-blok13')
        assert blok13.conflicting == ['test-blok1']
        assert blok1.conflicting_by == ['test-blok13']

    def test_install_1by1(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok1',))
        with pytest.raises(RegistryConflictingException):
            registry.upgrade(install=('test-blok13',))

    def test_install_both(self, registry_testblok):
        registry = registry_testblok
        with pytest.raises(RegistryConflictingException):
            registry.upgrade(install=('test-blok1', 'test-blok13'))

    def test_uninstall_first_and_intall_another_in_two_step(
        self, registry_testblok
    ):
        registry = registry_testblok
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok13 = Blok.query().filter(Blok.name == 'test-blok13').first()
        registry.upgrade(install=('test-blok1',))
        assert testblok1.state == 'installed'
        assert testblok13.state == 'uninstalled'
        registry.upgrade(uninstall=('test-blok1',))
        assert testblok1.state == 'uninstalled'
        assert testblok13.state == 'uninstalled'
        registry.upgrade(install=('test-blok13',))
        assert testblok1.state == 'uninstalled'
        assert testblok13.state == 'installed'

    def test_replace_blok_by_another_in_one_step(self, registry_testblok):
        registry = registry_testblok
        Blok = registry.System.Blok
        testblok1 = Blok.query().filter(Blok.name == 'test-blok1').first()
        testblok13 = Blok.query().filter(Blok.name == 'test-blok13').first()
        registry.upgrade(install=('test-blok1',))
        assert testblok1.state == 'installed'
        assert testblok13.state == 'uninstalled'
        registry.upgrade(uninstall=('test-blok1',), install=('test-blok13',))
        assert testblok1.state == 'uninstalled'
        assert testblok13.state == 'installed'


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                    reason='Not for MySQL and MariaDB')
class TestBlokUndefined:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok):
        transaction = registry_testblok.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_blok_15_exist(self, registry_testblok):
        registry = registry_testblok
        Blok = registry.System.Blok
        query = Blok.query().filter(Blok.name == 'test-blok15')
        if not query.count():
            pytest.fail('No blok found')

        testblok15 = query.first()
        assert testblok15.state == 'uninstalled'
        assert testblok15.version == '1.0.0'
        assert testblok15.installed_version is None
        assert testblok15.short_description == 'Test blok15'

    def test_blok_undefined_exist(self, registry_testblok):
        registry = registry_testblok
        Blok = registry.System.Blok
        query = Blok.query().filter(Blok.name == 'test-undefined')
        if not query.count():
            pytest.fail('No blok found')

        testblokundefined = query.first()
        assert testblokundefined.state == 'undefined'
        assert testblokundefined.version == '0.0.0'
        assert testblokundefined.installed_version is None
        assert testblokundefined.short_description == 'Blok undefined'

    def test_install_both(self, registry_testblok):
        registry = registry_testblok
        with pytest.raises(RegistryException):
            registry.upgrade(install=('test-blok15',))

    def test_install_undefined(self, registry_testblok):
        registry = registry_testblok
        with pytest.raises(RegistryException):
            registry.upgrade(install=('test-undefined',))


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                    reason='Not for MySQL and MariaDB')
class TestBlokOrder:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok):
        transaction = registry_testblok.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def check_order(self, registry, mode, wanted):
        Test = registry.Test
        bloks = Test.query().filter(Test.mode == mode).all().blok
        for _wanted in wanted:
            assert _wanted in bloks

    def test_install(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok3',))
        self.check_order(registry, 'install', [
            'test-blok1', 'test-blok2', 'test-blok3'])

    def test_uninstall(self, registry_testblok):
        from anyblok.blok import Blok, BlokManager
        old_uninstall = Blok.uninstall

        uninstalled = []

        def uninstall(self):
            cls = self.__class__
            uninstalled.extend(
                [x for x, y in BlokManager.bloks.items() if y is cls])

        try:
            Blok.uninstall = uninstall
            registry = registry_testblok
            registry.upgrade(install=('test-blok3',))
            registry.upgrade(uninstall=('test-blok1',))
            assert uninstalled == [
                'test-blok3', 'test-blok2', 'test-blok1']
        finally:
            Blok.uninstall = old_uninstall

    def test_update(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok3',))
        registry.upgrade(update=('test-blok1',))
        self.check_order(registry, 'update', [
            'test-blok1', 'test-blok2', 'test-blok3'])

    def test_load(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok3',))
        self.check_order(registry, 'load', [
            'anyblok-core', 'test-blok1', 'test-blok2', 'test-blok3'])


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                    reason='Not for MySQL and MariaDB')
class TestBlokModel:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok):
        transaction = registry_testblok.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_remove_foreign_key_after_uninstallation(self, registry_testblok):
        registry = registry_testblok
        registry.upgrade(install=('test-blok7', 'test-blok8'))
        t2 = registry.Test2.insert(label="test2")
        registry.Test.insert(label="Test1", test2=t2.id)
        from sqlalchemy.exc import IntegrityError
        with pytest.raises(IntegrityError):
            with registry.begin_nested():
                registry.Test2.execute_sql_statement(
                    registry.Test2.delete_sql_statement())

        registry.upgrade(uninstall=('test-blok8',))
        registry.Test2.execute_sql_statement(
            registry.Test2.delete_sql_statement())

    def test_auto_migration_is_between_pre_and_post_migration_1(
        self, registry_testblok
    ):
        registry = registry_testblok
        registry.upgrade(install=('test-blok14',))
        blok = BlokManager.get('test-blok14')
        assert blok.table_exist_before_automigration is False
        assert blok.table_exist_after_automigration is True
        registry.Test.insert()
        assert registry.Test.query().count() == 1

    def test_auto_migration_is_between_pre_and_post_migration_2(
        self, registry_testblok
    ):
        registry = registry_testblok
        registry.upgrade(install=('test-blok14',))
        blok = BlokManager.get('test-blok14')
        registry.Test.insert()
        assert registry.Test.query().count() == 1
        registry.upgrade(update=('test-blok14',))
        assert blok.table_exist_before_automigration is True
        assert blok.table_exist_after_automigration is True
        assert registry.Test.query().count() == 1


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                    reason='Not for MySQL and MariaDB')
class TestBlokSession:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok_func):
        request.addfinalizer(
            registry_testblok_func.unittest_transaction.rollback)

    def test_session_with_no_change(self, registry_testblok_func):
        registry = registry_testblok_func
        Session = registry.Session
        registry.upgrade(install=('test-blok1',))
        assert Session is registry.Session

    def test_session_with_change_query(self, registry_testblok_func):
        registry = registry_testblok_func
        Session = registry.Session
        registry.upgrade(install=('test-blok11',))
        assert Session is not registry.Session

    def test_session_with_change_session(self, registry_testblok_func):
        registry = registry_testblok_func
        Session = registry.Session
        registry.upgrade(install=('test-blok12',))
        assert Session is not registry.Session


@pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                    reason='Not for MySQL and MariaDB')
class TestBlokInstallLifeCycle:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_testblok_func):
        def clear_called_methods():
            BlokManager.get('test-blok16').called_methods = []
        request.addfinalizer(clear_called_methods)
        request.addfinalizer(
            registry_testblok_func.unittest_transaction.rollback)

    @pytest.fixture()
    def registry_blok16_installed(self, registry_testblok_func):
        registry_testblok_func.upgrade(install=('test-blok16',))
        blok = BlokManager.get('test-blok16')
        blok.called_methods = []
        return registry_testblok_func

    def test_install_without_demo(self, registry_testblok_func):
        registry = registry_testblok_func
        registry.System.Parameter.set("with-demo", False)
        registry.upgrade(install=('test-blok16',))
        blok = BlokManager.get('test-blok16')
        assert blok.called_methods == [
            "pre_migration", "post_migration", "update",
        ]

    def test_install_with_demo(self, registry_testblok_func):
        registry = registry_testblok_func
        registry.System.Parameter.set("with-demo", True)
        registry.upgrade(install=('test-blok16',))
        blok = BlokManager.get('test-blok16')
        assert blok.called_methods == [
            "pre_migration", "post_migration", "update", "update_demo",
        ]

    def test_update_without_demo(self, registry_blok16_installed):
        registry = registry_blok16_installed
        registry.System.Parameter.set("with-demo", False)
        registry.upgrade(update=('test-blok16',))
        blok = BlokManager.get('test-blok16')
        assert blok.called_methods == [
            "pre_migration", "post_migration", "update",
        ]

    def test_update_with_demo(self, registry_blok16_installed):
        registry = registry_blok16_installed
        registry.System.Parameter.set("with-demo", True)
        registry.upgrade(update=('test-blok16',))
        blok = BlokManager.get('test-blok16')
        assert blok.called_methods == [
            "pre_migration", "post_migration", "update", "update_demo",
        ]

    def test_unistall_without_demo(self, registry_blok16_installed):
        registry = registry_blok16_installed
        registry.System.Parameter.set("with-demo", False)
        registry.upgrade(uninstall=('test-blok16',))
        blok = BlokManager.get('test-blok16')
        assert blok.called_methods == [
            "uninstall",
        ]

    def test_unistall_with_demo(self, registry_blok16_installed):
        registry = registry_blok16_installed
        registry.System.Parameter.set("with-demo", True)
        registry.upgrade(uninstall=('test-blok16',))
        blok = BlokManager.get('test-blok16')
        assert blok.called_methods == [
            "uninstall_demo", "uninstall",
        ]

    def test_update_loadwithoutmigration(self, registry_blok16_installed):
        registry = registry_blok16_installed
        registry.loadwithoutmigration = True
        registry.System.Parameter.set("with-demo", True)
        registry.upgrade(update=('test-blok16',))
        blok = BlokManager.get('test-blok16')
        assert blok.called_methods == []

    def test_update_withoutautomigration(self, registry_blok16_installed):
        registry = registry_blok16_installed
        registry.withoutautomigration = True
        registry.System.Parameter.set("with-demo", True)
        registry.upgrade(update=('test-blok16',))
        blok = BlokManager.get('test-blok16')
        assert blok.called_methods == [
            "pre_migration", "post_migration", "update", "update_demo",
        ]



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok import Declarations
from anyblok.config import Configuration
from anyblok.column import Integer, String
from anyblok.relationship import One2Many, ordering_list
from .conftest import init_registry, reset_db


register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin


@pytest.fixture(
    scope="class",
    params=[
        ('prefix', 'suffix'),
        ('', ''),
    ]
)
def db_schema(request, bloks_loaded):
    Configuration.set('prefix_db_schema', request.param[0])
    Configuration.set('suffix_db_schema', request.param[1])

    def rollback():
        Configuration.set('prefix_db_schema', '')
        Configuration.set('suffix_db_schema', '')

    request.addfinalizer(rollback)


def _complete_one2many(**kwargs):
    primaryjoin = "ModelAddress.id == ModelPerson.address_id"

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address_id = Integer(foreign_key=Model.Address.use('id'))

    @register(Model)  # noqa
    class Address:

        persons = One2Many(model=Model.Person,
                           remote_columns="address_id",
                           primaryjoin=primaryjoin,
                           many2one="address")


def _complete_one2many_with_orderlist(**kwargs):
    primaryjoin = "ModelAddress.id == ModelPerson.address_id"

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address_id = Integer(foreign_key=Model.Address.use('id'))

    @register(Model)  # noqa
    class Address:

        persons = One2Many(model=Model.Person,
                           remote_columns="address_id",
                           primaryjoin=primaryjoin,
                           order_by="ModelPerson.name",
                           collection_class=ordering_list('name'),
                           many2one="address")


def _complete_one2many_with_schema(**kwargs):
    primaryjoin = "ModelAddress.id == ModelPerson.address_id"

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2m_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:
        __db_schema__ = 'test_db_m2m_schema'

        name = String(primary_key=True)
        address_id = Integer(foreign_key=Model.Address.use('id'))

    @register(Model)  # noqa
    class Address:
        __db_schema__ = 'test_db_m2m_schema'

        persons = One2Many(model=Model.Person,
                           remote_columns="address_id",
                           primaryjoin=primaryjoin,
                           many2one="address")


def _complete_one2many_with_diferent_schema1(**kwargs):
    primaryjoin = "ModelAddress.id == ModelPerson.address_id"

    @register(Model)
    class Address:
        __db_schema__ = 'test_other_db_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:
        __db_schema__ = 'test_db_m2m_schema'

        name = String(primary_key=True)
        address_id = Integer(foreign_key=Model.Address.use('id'))

    @register(Model)  # noqa
    class Address:
        __db_schema__ = 'test_other_db_schema'

        persons = One2Many(model=Model.Person,
                           remote_columns="address_id",
                           primaryjoin=primaryjoin,
                           many2one="address")


def _complete_one2many_with_diferent_schema2(**kwargs):
    primaryjoin = "ModelAddress.id == ModelPerson.address_id"

    @register(Model)
    class Address:
        __db_schema__ = 'test_other_db_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address_id = Integer(foreign_key=Model.Address.use('id'))

    @register(Model)  # noqa
    class Address:
        __db_schema__ = 'test_other_db_schema'

        persons = One2Many(model=Model.Person,
                           remote_columns="address_id",
                           primaryjoin=primaryjoin,
                           many2one="address")


@pytest.fixture(
    scope="class",
    params=[
        _complete_one2many,
        _complete_one2many_with_orderlist,
        _complete_one2many_with_schema,
        _complete_one2many_with_diferent_schema1,
        _complete_one2many_with_diferent_schema2,
    ]
)
def registry_complete_one2many(request, bloks_loaded, db_schema):
    reset_db()
    registry = init_registry(request.param)
    request.addfinalizer(registry.close)
    return registry


@pytest.mark.relationship
class TestCompleteOne2Many:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_complete_one2many):
        transaction = registry_complete_one2many.begin_nested()
        request.addfinalizer(transaction.rollback)

    def test_complete_one2many(self, registry_complete_one2many):
        registry = registry_complete_one2many

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")
        address.persons.append(person)

        assert person.address is address

    def test_one2many_autodoc(self, registry_complete_one2many):
        registry = registry_complete_one2many
        registry.loaded_namespaces_first_step['Model.Address'][
            'persons'].autodoc_get_properties()

    def test_complete_one2many_expire_field(self, registry_complete_one2many):
        registry = registry_complete_one2many
        assert ('address',) in registry.expire_attributes[
            'Model.Person']['address_id']
        assert ('address', 'persons') in registry.expire_attributes[
            'Model.Person']['address_id']

    def test_get_field_description(self, registry_complete_one2many):
        registry = registry_complete_one2many
        assert registry.Address.fields_description('persons') == {
            'persons': {
                'id': 'persons',
                'label': 'Persons',
                'local_columns': ['id'],
                'model': 'Model.Person',
                'nullable': True,
                'primary_key': False,
                'remote_columns': ['address_id'],
                'remote_name': 'address',
                'type': 'One2Many'
            },
        }


def _multi_fk_one2many():

    @register(Model)
    class Test:

        id = Integer(primary_key=True, unique=True)
        id2 = String(primary_key=True, unique=True)

    @register(Model)
    class Test2:

        id = Integer(primary_key=True)
        test_id = Integer(foreign_key=Model.Test.use('id'))
        test_id2 = String(foreign_key=Model.Test.use('id2'))

    @register(Model)  # noqa
    class Test:

        test2 = One2Many(model=Model.Test2, many2one="test")


@pytest.fixture(scope="class")
def registry_multi_fk_one2many(request, bloks_loaded):
    reset_db()
    registry = init_registry(_multi_fk_one2many)
    request.addfinalizer(registry.close)
    return registry


@pytest.mark.relationship
class TestMultiFkOne2Many:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_multi_fk_one2many):
        transaction = registry_multi_fk_one2many.begin_nested()
        request.addfinalizer(transaction.rollback)

    def test_with_multi_foreign_key(self, registry_multi_fk_one2many):
        registry = registry_multi_fk_one2many
        t1 = registry.Test.insert(id2="test")
        t2 = registry.Test2.insert(test=t1)
        assert len(t1.test2) == 1
        assert t1.test2[0] is t2

    def test_with_multi_foreign_key_expire_field(
        self, registry_multi_fk_one2many
    ):
        registry = registry_multi_fk_one2many
        assert ('test',) in registry.expire_attributes[
            'Model.Test2']['test_id']
        assert ('test', 'test2') in registry.expire_attributes[
            'Model.Test2']['test_id']
        assert ('test',) in registry.expire_attributes[
            'Model.Test2']['test_id2']
        assert ('test', 'test2') in registry.expire_attributes[
            'Model.Test2']['test_id2']


def _minimum_one2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()
        persons = One2Many(model='Model.Person')

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address_id = Integer(foreign_key=Model.Address.use('id'))


def one2many_on_mapper(**kwargs):

    @register(Model, tablename="x")
    class Address:

        id = Integer(primary_key=True, db_column_name="x1")
        street = String(db_column_name="x2")
        zip = String(db_column_name="x3")
        city = String(db_column_name="x4")
        persons = One2Many(model='Model.Person')

    @register(Model, tablename="y")
    class Person:

        name = String(primary_key=True, db_column_name="y1")
        address_id = Integer(foreign_key=Model.Address.use('id'),
                             db_column_name="y2")


def _minimum_one2many_with_schema(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2m_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()
        persons = One2Many(model='Model.Person')

    @register(Model)
    class Person:
        __db_schema__ = 'test_db_m2m_schema'

        name = String(primary_key=True)
        address_id = Integer(foreign_key=Model.Address.use('id'))


def _minimum_one2many_remote_field_in_mixin(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Mixin)
    class MPerson:
        address_id = Integer(foreign_key=Model.Address.use('id'))

    @register(Model)
    class Person(Mixin.MPerson):

        name = String(primary_key=True)

    @register(Model)  # noqa
    class Address:

        persons = One2Many(model=Model.Person)


def _one2many_with_str_model(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        address_id = Integer(foreign_key=Model.Address.use('id'))

    @register(Model)  # noqa
    class Address:

        persons = One2Many(model='Model.Person')


@pytest.mark.relationship
class TestOne2Many:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        reset_db()
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_minimum_one2many(self):
        registry = self.init_registry(_minimum_one2many)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")
        address.persons.append(person)

    def test_minimum_one2many_on_mapper(self):
        registry = self.init_registry(one2many_on_mapper)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")
        address.persons.append(person)

    def test_minimum_one2many_with_schema(self, db_schema):
        registry = self.init_registry(_minimum_one2many_with_schema)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")
        address.persons.append(person)

    def test_minimum_one2many_remote_field_in_mixin(self):
        registry = self.init_registry(_minimum_one2many_remote_field_in_mixin)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")
        address.persons.append(person)

    def test_one2many_with_str_model(self):
        registry = self.init_registry(_one2many_with_str_model)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")
        address.persons.append(person)

    def test_same_model_backref(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                parent_id = Integer(foreign_key='Model.Test=>id')
                children = One2Many(model='Model.Test', many2one='parent')

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert(parent=t1)
        assert t1.children[0] is t2
        assert t2.parent is t1

    def test_complet_with_multi_foreign_key(self):

        def add_in_registry():
            primaryjoin = "ModelTest.id == ModelTest2.test_id and "
            primaryjoin += "ModelTest.id2 == ModelTest2.test_id2"

            @register(Model)
            class Test:

                id = Integer(primary_key=True, unique=True)
                id2 = String(primary_key=True, unique=True)

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                test_id = Integer(foreign_key=Model.Test.use('id'))
                test_id2 = String(foreign_key=Model.Test.use('id2'))

            @register(Model)  # noqa
            class Test:

                test2 = One2Many(model=Model.Test2,
                                 remote_columns=['test_id', 'test_id2'],
                                 primaryjoin=primaryjoin,
                                 many2one="test")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test")
        t2 = registry.Test2.insert(test=t1)
        assert len(t1.test2) == 1
        assert t1.test2[0] is t2

    def test_with_multi_parallel_foreign_key_auto_detect(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address_1_id = Integer(foreign_key=Model.Address.use('id'))
                address_2_id = Integer(foreign_key=Model.Address.use('id'))

            @register(Model)  # noqa
            class Address:

                persons = One2Many(model=Model.Person)

        registry = self.init_registry(add_in_registry)
        address_1 = registry.Address.insert()
        address_2 = registry.Address.insert()
        person = registry.Person.insert(
            name='test', address_1_id=address_1.id, address_2_id=address_2.id)
        assert address_1.persons == [person]
        assert address_2.persons == [person]

    def test_with_multi_parallel_foreign_key_with_specific_primaryjoin(self):

        def add_in_registry():
            primaryjoin = (
                "or_("
                "ModelAddress.id == ModelPerson.address_1_id,"
                "ModelAddress.id == ModelPerson.address_2_id"
                ")"
            )

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class Person:

                name = String(primary_key=True)
                address_1_id = Integer(foreign_key=Model.Address.use('id'))
                address_2_id = Integer(foreign_key=Model.Address.use('id'))

            @register(Model)  # noqa
            class Address:

                persons = One2Many(model=Model.Person,
                                   primaryjoin=primaryjoin)

        registry = self.init_registry(add_in_registry)
        address_1 = registry.Address.insert()
        address_2 = registry.Address.insert()
        person = registry.Person.insert(
            name='test', address_1_id=address_1.id, address_2_id=address_2.id)
        assert address_1.persons == [person]
        assert address_2.persons == [person]

    def test_order_list(self):

        def add_in_registry():
            primaryjoin = "ModelAddress.id == ModelPerson.address_id"

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class Person:

                id = Integer(primary_key=True)
                name = String()
                address_id = Integer(foreign_key=Model.Address.use('id'))

            @register(Model)  # noqa
            class Address:

                persons = One2Many(model=Model.Person,
                                   remote_columns="address_id",
                                   primaryjoin=primaryjoin,
                                   order_by="ModelPerson.name",
                                   collection_class=ordering_list('name'),
                                   many2one="address")

        registry = self.init_registry(add_in_registry)

        address = registry.Address.insert()
        assert address.persons.ordering_attr == 'name'

        p3 = registry.Person.insert(name='test 3', address_id=address.id)
        p1 = registry.Person.insert(name='test 1', address_id=address.id)
        p4 = registry.Person.insert(name='test 4', address_id=address.id)
        p2 = registry.Person.insert(name='test 2', address_id=address.id)
        address.refresh()
        assert address.persons == [p1, p2, p3, p4]



import pytest


@pytest.mark.skip_unless_demo_data_installed
def test_skip_without_demo_rollback_registry(rollback_registry):
    # If Anyblok test case database is setup without `--with-demo` so this
    # test should be skipped otherwise `with-demo` parameter should be True
    assert rollback_registry.System.Parameter.get("with-demo", False) is True, (
        "`--with-demo` is not used to setup database to run AnyBlok tests "
        "so this test is intentionally skipped to validate "
        "skip_unless_demo_data_installed pytest marker behavior"
    )


@pytest.mark.skip_while_demo_data_installed
def test_skip_with_demo_rollback_registry(rollback_registry):
    # If Anyblok test case database is setup with `--with-demo` so this
    # test should be skipped otherwise `with-demo` parameter should be False
    # skip_unless_demo_data_installed marker
    assert rollback_registry.System.Parameter.get(
        "with-demo", False
    ) is False, (
        "`--with-demo` is not used to setup database to run AnyBlok tests "
        "so this test is intentionally skipped to validate "
        "skip_unless_demo_data_installed pytest marker behavior"
    )



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2021 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok
from anyblok.common import sgdb_in


class TestBlok(Blok):

    version = '1.0.0'
    table_exist_before_automigration = None
    table_exist_after_automigration = None

    @classmethod
    def import_declaration_module(cls):
        from . import test  # noqa

    @classmethod
    def reload_declaration_module(cls, reload):
        from . import test
        reload(test)

    def is_table_exist(self):
        engine = self.anyblok.engine
        if sgdb_in(engine, ['PostgreSQL']):
            return bool(self.anyblok.execute("""
                select count(*)
                from pg_catalog.pg_tables
                where tablename = 'test';
            """).fetchone()[0])
        elif sgdb_in(engine, ['MySQL']):
            return bool(self.anyblok.execute("""
                show tables like 'test';
            """).fetchall())
        elif sgdb_in(engine, ['MsSQL']):
            query = """
                SELECT count(*)
                FROM INFORMATION_SCHEMA.TABLES
                WHERE TABLE_TYPE = 'BASE TABLE'
                      AND TABLE_CATALOG='%s'
                      AND TABLE_NAME='test'
            """ % self.anyblok.db_name
            return bool(self.anyblok.execute(query).fetchall()[0][0])

    def pre_migration(self, latest_version):
        self.__class__.table_exist_before_automigration = self.is_table_exist()

    def post_migration(self, latest_version):
        self.__class__.table_exist_after_automigration = self.is_table_exist()



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.blok import Blok
from anyblok.release import version
from logging import getLogger
logger = getLogger(__name__)


class AnyBlokCore(Blok):
    """This Blok is required in all AnyBlok applications.

    This Blok provides the main fonctionalities for Bloks management (install,
    update, uninstall…).

    It also brings the representation of Anyblok objects (Models, Fields, etc.)
    within the database itself, and some fundamental facilities.

    * Core Models

      These are pure code Models, used as base classes:

      - Base: inherited by all Models
      - SqlBase: inherited by all models backed by an SQL table
      - SqlViewBase: inherited by all models bacled by an SQL view

    * System Models

      These correspond to actual tables in the table. They provide reflection
      or fundamental facilities.

      - Blok: represent all *available* Bloks, with their state and more
      - Model
      - Field
      - Column
      - Relationship
      - :class:`Sequence <.system.sequence.Sequence>`: database sequences,
        for use in applications.
      - :class:`Parameter <.system.parameter.Parameter>`: application
        parameters
    """

    version = version
    autoinstall = True
    priority = 0
    author = 'Suzanne Jean-Sébastien'
    logo = '../anyblok-logo_alpha_256.png'

    def pre_migration(self, latest_version):  # pragma: no cover
        if latest_version is None:
            return

        if latest_version < '0.4.1':
            self.pre_migration_0_4_1_fields_become_polymorphic(latest_version)

    def pre_migration_0_4_1_fields_become_polymorphic(
        self, latest_version
    ):  # pragma: no cover
        logger.info("Pre Migration %s => %s: Field, Column, Relation Ship "
                    "become prolymophic models" % (latest_version,
                                                   self.version))
        system_field = self.anyblok.migration.table('system_field')
        system_field.column().add(self.anyblok.System.Field.entity_type)
        self.anyblok.execute(
            "UPDATE system_field SET entity_type='Model.System.Field'")
        query = """
            INSERT INTO system_field (
                name,
                model,
                code,
                label,
                ftype,
                entity_type)
            SELECT
                name,
                model,
                code,
                label,
                ftype,
                '%(entity_type)s' AS entity_type
            FROM %(table)s
        """
        self.anyblok.execute(
            query % {'entity_type': 'Model.System.Column',
                     'table': 'system_column'})
        self.anyblok.execute(
            query % {'entity_type': 'Model.System.RelationShip',
                     'table': 'system_relationship'})
        system_column = self.anyblok.migration.table('system_column')
        system_column.column('code').drop()
        system_column.column('ftype').drop()
        system_column.column('label').drop()
        system_relationship = self.anyblok.migration.table(
            'system_relationship')
        system_relationship.column('code').drop()
        system_relationship.column('ftype').drop()
        system_relationship.column('label').drop()

    @classmethod
    def import_declaration_module(cls):
        from . import core  # noqa
        from . import system  # noqa
        from . import authorization  # noqa
        from . import documentation  # noqa

    @classmethod
    def reload_declaration_module(cls, reload):
        from . import core
        reload(core)
        from . import system
        reload(system)
        from . import authorization
        reload(authorization)
        from . import documentation
        reload(documentation)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.registry import RegistryManager
from anyblok import Declarations
from anyblok.environment import EnvironmentManager
register = Declarations.register
unregister = Declarations.unregister
Mixin = Declarations.Mixin


class OneInterface:
    pass


class TestCoreInterfaceMixin:

    @pytest.fixture(scope="class", autouse=True)
    def init_env(self, request):

        def revert():
            EnvironmentManager.set('current_blok', None)
            del RegistryManager.loaded_bloks['testMixin']

        request.addfinalizer(revert)
        RegistryManager.init_blok('testMixin')
        EnvironmentManager.set('current_blok', 'testMixin')

    @pytest.fixture(autouse=True)
    def init_blok(self):
        blokname = 'testMixin'
        RegistryManager.loaded_bloks[blokname]['Mixin'] = {
            'registry_names': []}

    def assertInMixin(self, *args):
        blokname = 'testMixin'
        blok = RegistryManager.loaded_bloks[blokname]
        assert len(blok['Mixin']['Mixin.MyMixin']['bases']) == len(args)
        for cls_ in args:
            has = cls_ in blok['Mixin']['Mixin.MyMixin']['bases']
            assert has is True

    def assertInRemoved(self, cls):
        core = RegistryManager.loaded_bloks['testMixin']['removed']
        if cls in core:
            return True

        self.fail('Not in removed')

    def test_add_interface(self):
        register(Mixin, cls_=OneInterface, name_='MyMixin')
        assert 'Mixin' == Mixin.MyMixin.__declaration_type__
        self.assertInMixin(OneInterface)
        dir(Declarations.Mixin.MyMixin)

    def test_add_interface_with_decorator(self):

        @register(Mixin)
        class MyMixin:
            pass

        assert 'Mixin' == Mixin.MyMixin.__declaration_type__
        self.assertInMixin(MyMixin)

    def test_add_two_interface(self):

        register(Mixin, cls_=OneInterface, name_="MyMixin")

        @register(Mixin)
        class MyMixin:
            pass

        self.assertInMixin(OneInterface, MyMixin)

    def test_remove_interface_with_1_cls_in_registry(self):

        register(Mixin, cls_=OneInterface, name_="MyMixin")
        self.assertInMixin(OneInterface)
        unregister(Mixin.MyMixin, OneInterface)

        blokname = 'testMixin'
        assert hasattr(Mixin, blokname) is False
        self.assertInMixin(OneInterface)
        self.assertInRemoved(OneInterface)

    def test_remove_interface_with_2_cls_in_registry(self):

        register(Mixin, cls_=OneInterface, name_="MyMixin")

        @register(Mixin)
        class MyMixin:
            pass

        self.assertInMixin(OneInterface, MyMixin)
        unregister(Mixin.MyMixin, OneInterface)
        self.assertInMixin(MyMixin, OneInterface)
        self.assertInRemoved(OneInterface)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.column import Integer, String, Selection
from anyblok.relationship import Many2One, One2One, Many2Many, One2Many
from anyblok.declarations import Declarations
from anyblok.bloks.anyblok_core.exceptions import SqlBaseException
from sqlalchemy.orm.exc import NoResultFound
from .conftest import init_registry


Model = Declarations.Model
register = Declarations.register


def declare_model():
    from anyblok import Declarations
    Model = Declarations.Model

    @Declarations.register(Model)
    class Test:
        id = Integer(primary_key=True)
        id2 = Integer()
        select = Selection(
            selections=[('key', 'value'), ('key2', 'value2')],
            default='key')


@pytest.fixture(scope="class")
def registry_declare_model(request, bloks_loaded):
    registry = init_registry(declare_model)
    request.addfinalizer(registry.close)
    return registry


class TestCoreSQLBase:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_declare_model):
        transaction = registry_declare_model.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_insert_and_query(self, registry_declare_model):
        registry = registry_declare_model
        t1 = registry.Test.insert(id2=1)
        assert registry.Test.query().first() == t1

    def test_insert_and_select(self, registry_declare_model):
        Test = registry_declare_model.Test
        t1 = Test.insert(id2=1)
        assert Test.execute_sql_statement(
            Test.select_sql_statement()).scalars().first() == t1

    def test_query_one_is_more_explicite(self, registry_declare_model):
        registry = registry_declare_model
        with pytest.raises(NoResultFound) as exc:
            registry.Test.query().one()

        assert (
            str(exc._excinfo[1]) ==
            "On Model 'Model.Test': No row was found when one was required")

    def test_query_dictone_is_more_explicite(self, registry_declare_model):
        registry = registry_declare_model
        with pytest.raises(NoResultFound) as exc:
            registry.Test.query().dictone()

        assert (
            str(exc._excinfo[1]) ==
            "On Model 'Model.Test': No row was found when one was required")

    def test_multi_insert(self, registry_declare_model):
        registry = registry_declare_model
        nb_value = 3
        registry.Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        assert registry.Test.query().count() == nb_value
        for x in range(nb_value):
            assert registry.Test.query().filter(
                registry.Test.id2 == x).count() == 1

    def test_classmethod_delete(self, registry_declare_model):
        registry = registry_declare_model
        nb_value = 3
        Test = registry.Test
        Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        assert Test.query().count() == nb_value
        id2 = 1
        Test.execute_sql_statement(
            Test.delete_sql_statement().where(Test.id2 == id2))
        assert registry.Test.query().count() == nb_value - 1
        assert registry.Test.query().filter(
            registry.Test.id2 != id2).count() == nb_value - 1

    def test_delete_by_query(self, registry_declare_model):
        registry = registry_declare_model
        nb_value = 3
        registry.Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        assert registry.Test.query().count() == nb_value
        t = registry.Test.query().first()
        id2 = t.id2
        t.delete(byquery=True)
        assert registry.Test.query().count() == nb_value - 1
        assert registry.Test.query().filter(
            registry.Test.id2 != id2).count() == nb_value - 1

    def test_delete(self, registry_declare_model):
        registry = registry_declare_model
        nb_value = 3
        registry.Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        assert registry.Test.query().count() == nb_value
        t = registry.Test.query().first()
        id2 = t.id2
        t.delete()
        assert registry.Test.query().count() == nb_value - 1
        assert registry.Test.query().filter(
            registry.Test.id2 != id2).count() == nb_value - 1

    def test_delete_with_get(self, registry_declare_model):
        registry = registry_declare_model
        nb_value = 3
        registry.Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        assert registry.Test.query().count() == nb_value
        t = registry.Test.query().first()
        assert registry.Test.query().get(t.id) is t
        t.delete()
        assert registry.Test.query().get(t.id) is None

    def test_expire(self, registry_declare_model):
        registry = registry_declare_model
        t = registry.Test.insert(id2=2)
        assert t.id2 == 2
        t.id2 = 3
        assert t.id2 == 3
        t.expire()
        assert t.id2 == 2

    def test_refresh(self, registry_declare_model):
        registry = registry_declare_model
        t = registry.Test.insert(id2=2)
        assert t.id2 == 2
        t.id2 = 3
        assert t.id2 == 3
        t.refresh()
        assert t.id2 == 2

    def test_modified(self, registry_declare_model):
        registry = registry_declare_model
        t = registry.Test.insert(id2=2)
        assert 'id2' not in t.get_modified_fields()
        t.flag_modified('id2')
        assert 'id2' in t.get_modified_fields()

    def test_classmethod_update(self, registry_declare_model):
        registry = registry_declare_model
        nb_value = 3
        Test = registry.Test
        Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        t = Test.query().filter_by(id2=1).one()
        assert Test.execute_sql_statement(
            Test.update_sql_statement().where(Test.id2 == 1).values(id2=100)
        ).rowcount == 1
        assert registry.Test.query().filter(
            registry.Test.id2 == 100).first() == t

    def test_update(self, registry_declare_model):
        registry = registry_declare_model
        nb_value = 3
        registry.Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        t = registry.Test.query().first()
        assert t.update(id2=100) == 1
        assert registry.Test.query().filter(
            registry.Test.id2 == 100).first() == t

    def test_update_byquery(self, registry_declare_model):
        registry = registry_declare_model
        nb_value = 3
        registry.Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        t = registry.Test.query().first()
        assert t.update(byquery=True, id2=100) == 1
        assert registry.Test.query().filter(
            registry.Test.id2 == 100).first() == t

    def test_get_primary_keys(self, registry_declare_model):
        registry = registry_declare_model
        assert registry.Test.get_primary_keys() == ['id']

    def test_to_and_from_primary_keys(self, registry_declare_model):
        registry = registry_declare_model
        nb_value = 3
        registry.Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        t = registry.Test.query().first()
        assert t.to_primary_keys() == {'id': t.id}
        assert registry.Test.from_primary_keys(id=t.id) == t

    def test_expire_with_column_selection(self, registry_declare_model):
        registry = registry_declare_model
        t = registry.Test.insert()
        t.select = 'key2'
        t.expire('select')
        assert t.select == 'key'

    def test_with_subquery(self, registry_declare_model):
        registry = registry_declare_model
        Test = registry.Test
        subquery = Test.query(Test.id2).subquery()
        assert subquery.c.keys() == ['id2']


def declare_model_with_m2o():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Test2:
        id = Integer(primary_key=True)
        name = String()
        test = Many2One(model=Model.Test, one2many="test2")


@pytest.fixture(scope="class")
def registry_declare_model_with_m2o(request, bloks_loaded):
    registry = init_registry(declare_model_with_m2o)
    request.addfinalizer(registry.close)
    return registry


class TestCoreSQLBaseM2O:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_declare_model_with_m2o):
        transaction = registry_declare_model_with_m2o.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_delete_entry_added_in_relationship(
        self, registry_declare_model_with_m2o
    ):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        assert len(t1.test2) == 1
        t2.delete()
        assert len(t1.test2) == 0

    def test_with_subquery_1(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        Test2 = registry.Test2
        subquery = Test2.query(Test2.test).subquery()
        assert subquery.c.keys() == ['test']

    def test_with_subquery_2(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        Test2 = registry.Test2
        subquery = Test2.query(Test2.test_id).subquery()
        assert subquery.c.keys() == ['test_id']

    def test_to_dict_m2o_with_pks(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        assert t1.to_dict('name') == {'name': 't1'}
        assert t2.to_dict('name') == {'name': 't2'}
        assert t2.to_dict('name', 'test') == {
            'name': 't2', 'test': {'id': t1.id}}

    def test_to_dict_o2m_with_pks(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        assert t1.to_dict('name', 'test2') == {
            'name': 't1', 'test2': [{'id': t2.id}]}

    def test_to_dict_m2o_with_column(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        assert t2.to_dict('name', ('test', ('name',))) == {
            'name': 't2', 'test': {'name': 't1'}}

    def test_to_dict_o2m_with_column(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        registry.Test2.insert(name='t2', test=t1)
        assert t1.to_dict('name', ('test2', ('name',))) == {
            'name': 't1', 'test2': [{'name': 't2'}]}

    def test_to_dict_m2o_with_all_columns(
        self, registry_declare_model_with_m2o
    ):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        assert t2.to_dict('name', ('test',)) == {
            'name': 't2', 'test': {'name': 't1',
                                   'id': t1.id,
                                   'test2': [{'id': t2.id}]}}
        assert t2.to_dict('name', ('test', None)) == {
            'name': 't2', 'test': {'name': 't1',
                                   'id': t1.id,
                                   'test2': [{'id': t2.id}]}}
        assert t2.to_dict('name', ('test', ())) == {
            'name': 't2', 'test': {'name': 't1',
                                   'id': t1.id,
                                   'test2': [{'id': t2.id}]}}

    def test_to_dict_o2m_with_all_columns(
        self, registry_declare_model_with_m2o
    ):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        assert t1.to_dict('name', ('test2',)) == {
            'name': 't1', 'test2': [{'name': 't2',
                                     'id': t2.id,
                                     'test_id': t1.id,
                                     'test': {'id': t1.id}}]}
        assert t1.to_dict('name', ('test2', None)) == {
            'name': 't1', 'test2': [{'name': 't2',
                                     'id': t2.id,
                                     'test_id': t1.id,
                                     'test': {'id': t1.id}}]}
        assert t1.to_dict('name', ('test2', ())) == {
            'name': 't1', 'test2': [{'name': 't2',
                                     'id': t2.id,
                                     'test_id': t1.id,
                                     'test': {'id': t1.id}}]}

    def test_bad_definition_of_relation(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        with pytest.raises(SqlBaseException):
            t2.to_dict('name', ('test', 'name'))

    def test_bad_definition_of_relation2(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        with pytest.raises(SqlBaseException):
            t2.to_dict('name', ('test', ('name',), ()))

    def test_bad_definition_of_relation3(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        with pytest.raises(SqlBaseException):
            t2.to_dict('name', ())

    def test_refresh_update_m2o(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test.insert(name='t3')
        assert t1.test2 == [t2]
        assert t2.test is t1
        assert t2.test_id == t1.id
        assert t3.test2 == []
        t2.test = t3
        assert t1.test2 == []
        assert t2.test is t3
        assert t2.test_id is t3.id
        assert t3.test2 == [t2]
        t2.test = None
        assert t1.test2 == []
        assert t2.test is None
        assert t2.test_id is None
        assert t3.test2 == []

    def test_refresh_update_m2o_2(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test.insert(name='t3')
        assert t1.test2 == [t2]
        assert t2.test is t1
        assert t2.test_id == t1.id
        assert t3.test2 == []
        t2.test_id = t3.id
        assert t1.test2 == []
        assert t2.test is t3
        assert t2.test_id == t3.id
        assert t3.test2 == [t2]

    def test_refresh_update_m2o_3(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test.insert(name='t3')
        assert t1.test2 == [t2]
        assert t2.test is t1
        assert t2.test_id == t1.id
        assert t3.test2 == []
        t3.test2.append(t2)
        assert t1.test2 == []
        assert t2.test is t3
        assert t2.test_id == t3.id
        assert t3.test2 == [t2]
        t3.test2.remove(t2)
        assert t1.test2 == []
        assert t2.test is None
        assert t2.test_id is None
        assert t3.test2 == []

    def test_find_relationship_by_relationship(
        self, registry_declare_model_with_m2o
    ):
        registry = registry_declare_model_with_m2o
        fields = registry.Test2.find_relationship('test')
        assert 'test' in fields
        assert 'test_id' in fields

    def test_find_relationship_by_column(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        fields = registry.Test2.find_relationship('test_id')
        assert 'test' in fields
        assert 'test_id' in fields

    def test_getFieldType(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        assert registry.Test.getFieldType('id') == 'Integer'
        assert registry.Test.getFieldType('name') == 'String'
        assert registry.Test2.getFieldType('id') == 'Integer'
        assert registry.Test2.getFieldType('name') == 'String'
        assert registry.Test2.getFieldType('test') == 'Many2One'
        assert registry.Test.getFieldType('test2') == 'One2Many'

    def test_repr_m2o(self, registry_declare_model_with_m2o):
        registry = registry_declare_model_with_m2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        wanted = (
            "<Model.Test2(id=%s, name='t2', test=<Model.Test(id=%s)>"
            ", test_id=%s)>" % (t2.id, t1.id, t1.id))
        assert repr(t2) == wanted


def declare_model_with_o2o():
    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Test2:
        id = Integer(primary_key=True)
        name = String()
        test = One2One(model=Model.Test, backref='test2')


@pytest.fixture(scope="class")
def registry_declare_model_with_o2o(request, bloks_loaded):
    registry = init_registry(declare_model_with_o2o)
    request.addfinalizer(registry.close)
    return registry


class TestCoreSQLBaseo2O:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_declare_model_with_o2o):
        transaction = registry_declare_model_with_o2o.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_to_dict_o2o_with_pks(self, registry_declare_model_with_o2o):
        registry = registry_declare_model_with_o2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        assert t1.to_dict('name') == {'name': 't1'}
        assert t2.to_dict('name') == {'name': 't2'}
        assert t2.to_dict('name', 'test') == {
            'name': 't2', 'test': {'id': t1.id}}
        assert t1.to_dict('name', 'test2') == {
            'name': 't1', 'test2': {'id': t2.id}}

    def test_to_dict_o2o_with_column(self, registry_declare_model_with_o2o):
        registry = registry_declare_model_with_o2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        assert t2.to_dict('name', ('test', ('name',))) == {
            'name': 't2', 'test': {'name': 't1'}}
        assert t1.to_dict('name', ('test2', ('name',))) == {
            'name': 't1', 'test2': {'name': 't2'}}

    def test_to_dict_o2o_with_all_columns(
        self, registry_declare_model_with_o2o
    ):
        registry = registry_declare_model_with_o2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        assert t2.to_dict('name', ('test',)) == {
            'name': 't2', 'test': {'name': 't1',
                                   'id': t1.id,
                                   'test2': {'id': t2.id}}}
        assert t2.to_dict('name', ('test', None)) == {
            'name': 't2', 'test': {'name': 't1',
                                   'id': t1.id,
                                   'test2': {'id': t2.id}}}
        assert t2.to_dict('name', ('test', ())) == {
            'name': 't2', 'test': {'name': 't1',
                                   'id': t1.id,
                                   'test2': {'id': t2.id}}}

    def test_refresh_update_o2o(self, registry_declare_model_with_o2o):
        registry = registry_declare_model_with_o2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test.insert(name='t3')
        assert t1.test2 is t2
        assert t2.test is t1
        assert t2.test_id is t1.id
        assert t3.test2 is None
        t2.test = t3
        assert t1.test2 is None
        assert t2.test is t3
        assert t2.test_id == t3.id
        assert t3.test2 is t2

    def test_refresh_update_o2o_2(self, registry_declare_model_with_o2o):
        registry = registry_declare_model_with_o2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test.insert(name='t3')
        assert t1.test2 is t2
        assert t2.test is t1
        assert t3.test2 is None
        t3.test2 = t2
        assert t1.test2 is None
        assert t2.test is t3
        assert t2.test_id is t3.id
        assert t3.test2 is t2

    def test_refresh_update_o2o_3(self, registry_declare_model_with_o2o):
        registry = registry_declare_model_with_o2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test.insert(name='t3')
        assert t1.test2 is t2
        assert t2.test is t1
        assert t2.test_id == t1.id
        assert t3.test2 is None
        t2.test_id = t3.id
        assert t1.test2 is None
        assert t2.test is t3
        assert t2.test_id == t3.id
        assert t3.test2 is t2

    def test_repr_o2o(self, registry_declare_model_with_o2o):
        registry = registry_declare_model_with_o2o
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        wanted = (
            "<Model.Test2(id=%s, name='t2', test=<Model.Test(id=%s)>"
            ", test_id=%s)>" % (t2.id, t1.id, t1.id))
        assert repr(t2) == wanted


def declare_model_with_o2m():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Test2:
        id = Integer(primary_key=True)
        name = String()
        test_id = Integer(foreign_key=Model.Test.use('id'))

    @register(Model)  # noqa
    class Test:
        test2 = One2Many(model=Model.Test2, many2one="test")


@pytest.fixture(scope="class")
def registry_declare_model_with_o2m(request, bloks_loaded):
    registry = init_registry(declare_model_with_o2m)
    request.addfinalizer(registry.close)
    return registry


class TestCoreSQLBaseo2m:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_declare_model_with_o2m):
        transaction = registry_declare_model_with_o2m.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_refresh_update_o2m(self, registry_declare_model_with_o2m):
        registry = registry_declare_model_with_o2m
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test2.insert(name='t3')
        assert t1.test2 == [t2]
        assert t2.test is t1
        assert t3.test is None
        t1.test2.append(t3)
        assert t2 in t1.test2
        assert t3 in t1.test2
        assert t2.test is t1
        assert t3.test is t1
        t1.test2.remove(t2)
        assert t1.test2 == [t3]
        assert t2.test is None
        assert t3.test is t1

    def test_refresh_update_o2m_2(self, registry_declare_model_with_o2m):
        registry = registry_declare_model_with_o2m
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test2.insert(name='t3')
        assert t1.test2 == [t2]
        assert t2.test is t1
        assert t3.test is None
        t3.test = t1
        assert t2 in t1.test2
        assert t3 in t1.test2
        assert t2.test is t1
        assert t3.test is t1
        assert t3.test_id is t1.id

    def test_refresh_update_o2m_3(self, registry_declare_model_with_o2m):
        registry = registry_declare_model_with_o2m
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test2.insert(name='t3')
        assert t1.test2 == [t2]
        assert t2.test is t1
        assert t3.test is None
        t3.test_id = t1.id
        assert t2 in t1.test2
        assert t3 in t1.test2
        assert t2.test is t1
        assert t3.test is t1

    def test_repr_o2m(self, registry_declare_model_with_o2m):
        registry = registry_declare_model_with_o2m
        t1 = registry.Test.insert(name='t1')
        registry.Test2.insert(name='t2', test=t1)
        wanted = "<Model.Test(id=%s, name='t1', test2=<not loaded>)>" % t1.id
        assert repr(t1) == wanted


def declare_model_with_m2m():
    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Test2:
        id = Integer(primary_key=True)
        name = String()
        test = Many2Many(model=Model.Test, many2many='test2')


@pytest.fixture(scope="class")
def registry_declare_model_with_m2m(request, bloks_loaded):
    registry = init_registry(declare_model_with_m2m)
    request.addfinalizer(registry.close)
    return registry


class TestCoreSQLBasem2m:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_declare_model_with_m2m):
        transaction = registry_declare_model_with_m2m.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_to_dict_m2m_with_pks(self, registry_declare_model_with_m2m):
        registry = registry_declare_model_with_m2m
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2')
        t2.test.append(t1)
        assert t1.to_dict('name') == {'name': 't1'}
        assert t2.to_dict('name') == {'name': 't2'}
        assert t2.to_dict('name', 'test') == {
            'name': 't2', 'test': [{'id': t1.id}]}
        assert t1.to_dict('name', 'test2') == {
            'name': 't1', 'test2': [{'id': t2.id}]}

    def test_to_dict_m2m_with_column(self, registry_declare_model_with_m2m):
        registry = registry_declare_model_with_m2m
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2')
        t2.test.append(t1)
        assert t2.to_dict('name', ('test', ('name',))) == {
            'name': 't2', 'test': [{'name': 't1'}]}
        assert t1.to_dict('name', ('test2', ('name',))) == {
            'name': 't1', 'test2': [{'name': 't2'}]}

    def test_to_dict_m2m_with_all_columns(
        self, registry_declare_model_with_m2m
    ):
        registry = registry_declare_model_with_m2m
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2')
        t2.test.append(t1)
        assert t2.to_dict('name', ('test',)) == {
            'name': 't2', 'test': [{'name': 't1',
                                    'id': t1.id,
                                    'test2': [{'id': t2.id}]}]}
        assert t2.to_dict('name', ('test', None)) == {
            'name': 't2', 'test': [{'name': 't1',
                                    'id': t1.id,
                                    'test2': [{'id': t2.id}]}]}
        assert t2.to_dict('name', ('test', ())) == {
            'name': 't2', 'test': [{'name': 't1',
                                    'id': t1.id,
                                    'test2': [{'id': t2.id}]}]}
        assert t1.to_dict('name', ('test2',)) == {
            'name': 't1', 'test2': [{'name': 't2',
                                     'id': t2.id,
                                     'test': [{'id': t1.id}]}]}
        assert t1.to_dict('name', ('test2', None)) == {
            'name': 't1', 'test2': [{'name': 't2',
                                     'id': t2.id,
                                     'test': [{'id': t1.id}]}]}
        assert t1.to_dict('name', ('test2', ())) == {
            'name': 't1', 'test2': [{'name': 't2',
                                     'id': t2.id,
                                     'test': [{'id': t1.id}]}]}

    def test_refresh_update_m2m(self, registry_declare_model_with_m2m):
        registry = registry_declare_model_with_m2m
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2')
        t3 = registry.Test2.insert(name='t3')
        assert t1.test2 == []
        assert t2.test == []
        assert t3.test == []
        t1.test2.append(t2)
        assert t1.test2 == [t2]
        assert t2.test == [t1]
        assert t3.test == []
        t1.test2.append(t3)
        assert t2 in t1.test2
        assert t3 in t1.test2
        assert t2.test == [t1]
        assert t3.test == [t1]
        t1.test2.remove(t2)
        assert t1.test2 == [t3]
        assert t2.test == []
        assert t3.test == [t1]

    def test_refresh_update_m2m_2(self, registry_declare_model_with_m2m):
        registry = registry_declare_model_with_m2m
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2')
        t3 = registry.Test2.insert(name='t3')
        assert t1.test2 == []
        assert t2.test == []
        assert t3.test == []
        t2.test.append(t1)
        assert t1.test2 == [t2]
        assert t2.test == [t1]
        assert t3.test == []
        t3.test.append(t1)
        assert t2 in t1.test2
        assert t3 in t1.test2
        assert t2.test == [t1]
        assert t3.test == [t1]
        t2.test.remove(t1)
        assert t1.test2 == [t3]
        assert t2.test == []
        assert t3.test == [t1]

    def test_repr_m2m(self, registry_declare_model_with_m2m):
        registry = registry_declare_model_with_m2m
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2')
        t2.test.append(t1)
        wanted = (
            "<Model.Test2(id=%s, name='t2', test=<Model.Test len(1)>)>" % t2.id
        )
        assert repr(t2) == wanted



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.blok import BlokManager
from sys import modules
from anyblok.imp import ImportManager, ImportManagerException
from os.path import join
from .mockblok import mockblok


tests_path = join('/'.join(__file__.split('/')[:-1]), 'mockblok')
with open(join(tests_path, 'mockfile.py'), 'r') as fp:
    initial_file = fp.read()


def python_version():
    import sys
    return '%d.%d' % (sys.version_info.major, sys.version_info.minor)


class TestImportManager:

    @pytest.fixture(autouse=True)
    def reset(self, request):
        def close():
            if '.mockblok.' in modules:
                mod_2_del = []
                for mod in modules.keys():
                    if '.mockblok.' in mod:
                        mod_2_del.append(mod)

                for mod in mod_2_del:
                    del modules[mod]

            del BlokManager.bloks['mockblok']
            BlokManager.ordered_bloks.remove('mockblok')

        request.addfinalizer(close)
        BlokManager.bloks['mockblok'] = mockblok
        BlokManager.ordered_bloks.append('mockblok')

    def test_has_blok(self):
        ImportManager.add('mockblok')
        assert ImportManager.has('mockblok') is True
        assert ImportManager.has('mockblok2') is False

    def test_get_blok(self):
        ImportManager.add('mockblok')
        ImportManager.get('mockblok')

    def test_get_unexisting_blok(self):
        try:
            ImportManager.get('mockblok2')
            self.fail('No watchdog for inexisting blok module')
        except ImportManagerException:
            pass

    def test_imports(self):
        blok = ImportManager.add('mockblok')
        blok.imports()
        from .mockblok.mockpackage import mockfile1, mockfile2
        from .mockblok.mockpackage import submockpackage
        from .mockblok import mockfile
        assert mockfile1.foo == 'bar'
        assert mockfile2.foo == 'bar'
        assert submockpackage.mockfile1.foo == 'bar'
        assert submockpackage.mockfile2.foo == 'bar'
        assert mockfile.foo == 'bar'

    def test_reload(self):
        blok = ImportManager.add('mockblok')
        blok.imports()
        from .mockblok.mockpackage import mockfile1, mockfile2
        from .mockblok.mockpackage import submockpackage
        from .mockblok import mockfile

        with open(join(tests_path, 'mockpackage', 'mockfile1.py'), 'w') as fp:
            fp.write("""foo = 'reload'""")

        with open(join(tests_path, 'mockpackage', 'submockpackage',
                       'mockfile2.py'), 'w') as fp:
            fp.write("""foo = 'reload'""")

        with open(join(tests_path, 'mockfile.py'), 'w') as fp:
            fp.write("""foo = 'reload'""")

        try:
            assert mockfile1.foo == 'bar'
            assert mockfile2.foo == 'bar'
            assert submockpackage.mockfile1.foo == 'bar'
            assert submockpackage.mockfile2.foo == 'bar'
            assert mockfile.foo == 'bar'
            blok.reload()
            assert mockfile1.foo == 'reload'
            assert mockfile2.foo == 'bar'
            assert submockpackage.mockfile1.foo == 'bar'
            assert submockpackage.mockfile2.foo == 'reload'
            assert mockfile.foo == 'reload'
        finally:
            with open(join(tests_path, 'mockfile.py'), 'w') as fp:
                fp.write(initial_file)

            with open(join(tests_path, 'mockpackage',
                           'mockfile1.py'), 'w') as fp:
                fp.write(initial_file)

            with open(join(tests_path, 'mockpackage', 'submockpackage',
                           'mockfile2.py'), 'w') as fp:
                fp.write(initial_file)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Denis VIVIÈS <dvivies@geoblink.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest


@pytest.mark.usefixtures('rollback_registry')
class TestQueryScope:
    def test_dictone(self, rollback_registry):
        registry = rollback_registry
        query = registry.System.Model.query().filter_by(
            name='Model.System.Blok')
        model = query.one()
        assert query.dictone() == {
            'name': model.name,
            'table': model.table,
            'schema': model.schema,
            'is_sql_model': model.is_sql_model,
            'description': model.description,
        }

    def test_dictone_on_some_column(self, rollback_registry):
        registry = rollback_registry
        query = registry.System.Model.query('name', 'table').filter(
            registry.System.Model.name == 'Model.System.Blok')
        model = query.one()
        assert query.dictone() == {
            'name': model.name,
            'table': model.table,
        }

    def test_dictone_on_some_column_with_label(self, rollback_registry):
        registry = rollback_registry
        M = registry.System.Model
        query = M.query(M.name.label('n2'), M.table.label('t2')).filter(
            registry.System.Model.name == 'Model.System.Blok')
        model = query.one()
        assert query.dictone() == {
            'n2': model.n2,
            't2': model.t2,
        }

    def test_dictfirst(self, rollback_registry):
        registry = rollback_registry
        query = registry.System.Model.query()
        model = query.first()
        assert query.dictfirst() == {
            'name': model.name,
            'table': model.table,
            'schema': model.schema,
            'is_sql_model': model.is_sql_model,
            'description': model.description,
        }

    def test_dictfirst_on_some_column(self, rollback_registry):
        registry = rollback_registry
        query = registry.System.Model.query('name', 'table')
        model = query.first()
        assert query.dictfirst() == {
            'name': model.name,
            'table': model.table,
        }

    def test_dictfirst_on_some_column_with_label(self, rollback_registry):
        registry = rollback_registry
        M = registry.System.Model
        query = M.query(M.name.label('n2'), M.table.label('t2'))
        model = query.first()
        assert query.dictfirst() == {
            'n2': model.n2,
            't2': model.t2,
        }

    def test_dictall(self, rollback_registry):
        registry = rollback_registry
        query = registry.System.Model.query().limit(2)
        models = query.all()

        def to_dict(model):
            return {
                'name': model.name,
                'table': model.table,
                'schema': model.schema,
                'is_sql_model': model.is_sql_model,
                'description': model.description,
            }

        dictall = query.dictall()
        for i in range(2):
            assert to_dict(models[i]) in dictall

    def test_dictall_on_some_column(self, rollback_registry):
        registry = rollback_registry
        query = registry.System.Model.query('name', 'table').limit(2)
        models = query.all()

        def to_dict(model):
            return {
                'name': model.name,
                'table': model.table,
            }

        dictall = query.dictall()
        for i in range(2):
            assert to_dict(models[i]) in dictall

    def test_dictall_on_some_column_with_label(self, rollback_registry):
        registry = rollback_registry
        M = registry.System.Model
        query = M.query(M.name.label('n2'), M.table.label('t2')).limit(2)
        models = query.all()

        def to_dict(model):
            return {
                'n2': model.n2,
                't2': model.t2,
            }

        dictall = query.dictall()
        for i in range(2):
            assert to_dict(models[i]) in dictall

    def test_get_with_dict_use_prefix(self, rollback_registry):
        registry = rollback_registry
        M = registry.System.Field
        entry = M.query().get({'name': 'name', 'model': 'Model.System.Blok'})
        assert entry is not None

    def test_get_with_kwargs(self, rollback_registry):
        registry = rollback_registry
        M = registry.System.Field
        entry = M.query().get(name='name', model='Model.System.Blok')
        assert entry is not None



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pkg_resources import iter_entry_points
from logging import getLogger

logger = getLogger(__name__)


def get_model_plugins(registry):
    res = []
    for i in iter_entry_points('anyblok.model.plugin'):
        logger.info('AnyBlok Load model plugin: %r' % i)
        res.append(i.load()(registry))

    return res


class ModelPluginBase:

    def __init__(self, registry):
        self.registry = registry

    # def initialisation_tranformation_properties(self, properties,
    #                                             transformation_properties):
    #     """ Initialise the transform properties

    #     :param properties: the properties declared in the model
    #     :param new_type_properties: param to add in a new base if need
    #     """

    # def declare_field(self, name, field, namespace, properties,
    #                   transformation_properties):
    #     """Declare a field in the model

    #     :param name: field name
    #     :param field: field instance
    #     :param namespace: the namespace of the model
    #     :param properties: the properties of the model
    #     :param transformation_properties: the transformation properties
    #     """

    # def transform_base_attribute(self, attr, method, namespace, base,
    #                              transformation_properties,
    #                              new_type_properties):
    #     """ transform the attribute for the final Model

    #     :param attr: attribute name
    #     :param method: method pointer of the attribute
    #     :param namespace: the namespace of the model
    #     :param base: One of the base of the model
    #     :param transformation_properties: the properties of the model
    #     :param new_type_properties: param to add in a new base if need
    #     """

    # def transform_base(self, namespace, base,
    #                    transformation_properties,
    #                    new_type_properties):
    #     """ transform the base for the final Model

    #     :param namespace: the namespace of the model
    #     :param base: One of the base of the model
    #     :param transformation_properties: the properties of the model
    #     :param new_type_properties: param to add in a new base if need
    #     """

    # def insert_in_bases(self, new_base, namespace, properties,
    #                     transformation_properties):
    #     """Insert in a base the overload

    #     :param new_base: the base to be put on front of all bases
    #     :param namespace: the namespace of the model
    #     :param properties: the properties declared in the model
    #     :param transformation_properties: the properties of the model
    #     """

    # def after_model_construction(self, base, namespace,
    #                              transformation_properties):
    #     """Do some action with the constructed Model

    #     :param base: the Model class
    #     :param namespace: the namespace of the model
    #     :param transformation_properties: the properties of the model
    #     """
